#!/usr/bin/perl -w

use File::Basename;

# amavis-blocked.pl
#
# create a report about blocked messages for a given recipient based
# on the corresponding lines out of the amavisd-new logfile
#
# version: v0.3
# author: uwe s. fuerst <uf@phi.net>
#
# history:
# v0.1 - 04/04/30 - first version
# v0.2 - 04/05/10 - added special feature (2)
# v0.3 - 04/11/07 - cleanup for posting at [AMaViS-user]
# v0.4 - 04/11/08 - changed "detection" of "Hits: -" - thanks Bill!
#
# usage:
# amavis-blocked.pl <pattern> [MMM [DD]]
# - <pattern>: usually an email address
# - MMM: month (Jan, Feb, ..., Dec)
# - DD:  day   (1, 2, ..., 31)

# special features:
#
# (1) passed messages:
# create a link from amavis-blocked.pl to amavis-passed.pl and the
# output of the report will change to the messages that passed thru
#
# (2) special pattern "mailaddr":
# $ amavis-blocked.pl mailaddr
# this will show all valid recipients/senders with blocked/passed
# messages found in the logfile

# *** START OF CONFIGURATION AREA ***

# where does the amavisd-new logfile live?
$LOGFILE = "/var/amavis/amavis.log";

# DSN cutoff level (see amavisd.conf: $sa_dsn_cutoff_level)
$DSN = 8;

# *** END OF CONFIGURATION AREA ***

# some presets
$MYNAME = basename($0);
$Month = $Day = $First = $Last = "";
$Pat = $P = $P1 = $P2 = $P3 = "";
$bhc = $bnc = $ic = $sc = $clc = 0;

# subroutines
sub Usage {
    print "\n$MYNAME: ERROR: @_\n\n";
    print "Usage:\n";
    print "$MYNAME <pattern> [MMM [DD]]\n";
    print "- <pattern>: usually an email address\n";
    print "- MMM......: month (Jan, Feb, ..., Dec)\n";
    print "- DD.......: day   (1, 2, ..., 31)\n\n";
    exit 1;
}

sub ShowMAddr {
    my $maddr = "";
    open(LOG, "<", $LOGFILE) or die "Couldn't open logfile $LOGFILE";
    # dunno how to do this in Perl :-(
    open(CMD, "| sort -u") or die "Couldn't open pipe to 'sort'";
    while(<LOG>) {
        next if not /\s$Pat\s.*\s->\s/;
        @line = split(/\s->\s/);
        if($P eq "+") {
            @line = split(/\sMessage-ID:\s/, $line[1]);
            $maddr = $line[0];
            $maddr =~ y/,/\n/; }
        else {
            @line = split(/\sMessage-ID:\s/, $line[1]);
            $maddr = $line[0];
            @line = split(/,\s/, $line[0]);
            $maddr .= $line[$#line];
            $maddr =~ y/,/\n/;
        }
        printf CMD lc($maddr);
    }
    close(CMD);
    close(LOG);
    exit;
}

# *** MAIN ***

# what's my name?
if($MYNAME =~ /blocked/) {
    $Pat = "Blocked"; $P = "+"; $P1 = "="; $P3 = "-"; }
else {
    $Pat = "Passed"; $P = "-"; $P2 = ".";
}

# check for command line arguments
$ARGV[0] ? $MAddr = $ARGV[0] : die Usage("Argument missing!");
if ($MAddr eq "mailaddr") { ShowMAddr(); };
$Month = $ARGV[1] if $ARGV[1];
if($ARGV[2]) {
    $Day = $ARGV[2];
    if($Day < 10) {
        $Day = "  $Day"; }
    else {
        $Day = " $Day";
    };
};

# open the logfile
open(LOG, "<", $LOGFILE) or die "Couldn't open logfile $LOGFILE";

# look for the corresponding entries and summarize the information
while(<LOG>) {
    next if not /^$Month$Day.*\s$Pat\s.*$MAddr/i;
    # enum the recipient(s)
    @rcpt = split(/\s->\s/);
    @rcpt = split(/,\sMessage-ID:\s/, $rcpt[1]);
    $rcpt[0] =~ y/<>//d;
    $rcpt = $rcpt[0];
    # enum the sender
    @frm = split(/>\s->\s/);
    @frm = split(/</, $frm[0]);
    $frm = $frm[1];
    # $frm =~ s/^\?@/unknown@/;
    # if(!$frm) { $frm = "unknown" };
    if(!$frm) { $frm = "?@?" };
    # now parse the rest of the line
    @line = split;
    # memorize the first and last parsed line
    $First = sprintf("%s %2d %s", $line[0], $line[1], $line[2]) if not $First;
    $Last = sprintf("%s %2d %s", $line[0], $line[1], $line[2]);
    # that's a hack to catch "Hits: -"
    if((/Hits: -$/ || /Hits: -,/)) {
        $P eq "+" ? $hits[1] = $DSN - 0.001 : $hits[1] = 0; }
    else {
        @hits = split(/\sHits:\s/);
        $hits[1] =~ s/,\s(.*)$//;
    }
    if($line[7] eq "BANNED") {
        @bname = split(/\),\s</, "@line");
        @bname = split(/N=/, $bname[0]);
        @bname = split(/,/, $bname[$#bname]);
        $bn[++$bnc] = sprintf("%s|%s [%s] -> %s", "$Last", $frm, $bname[0], $rcpt); }
    elsif($line[7] eq "INFECTED") {
        @vir = split(/INFECTED\s\(/, "@line");
        @vir = split(/\),\s/, $vir[1]);
        $i[++$ic] = sprintf("%s|%s [%s] -> %s", "$Last", $frm, $vir[0], $rcpt); }
    elsif($line[7] eq "BAD-HEADER,") {
        $bh[++$bhc] = sprintf("%s|%8.3f|%s|%s -> %s", "$Last", $hits[1],
                              ($hits[1] < $DSN) ? $P : '-', $frm, $rcpt); }
                              #($hits[1] < $DSN && $frm ne $rcpt) ? $P : '-', $frm, $rcpt); }
    elsif($line[7] eq "CLEAN,") {
        $cl[++$clc] = sprintf("%s|%8.3f|-|%s -> %s", "$Last", $hits[1], $frm, $rcpt); }
    else {
        $s[++$sc] = sprintf("%s|%8.3f|%s|%s -> %s", "$Last", $hits[1],
                            ($hits[1] < $DSN) ? $P : '-', $frm, $rcpt);
                            #($hits[1] < $DSN && $frm ne $rcpt) ? $P : '-', $frm, $rcpt);
    }
}

# close the logfile
close(LOG);

# nothing to report?
($bhc + $bnc + $ic + $sc + $clc) or exit;

# show what we have
print "AMaViS $Pat Messages Report\n";
print "$P1=============================\n\n\n";
print "Report Start..................: $First\n";
print "Report End....................: $Last\n\n";
print "$Pat INFECTED messages.....$P2: $ic\n" if $ic;
print "$Pat BAD-HEADER messages...$P2: $bhc\n" if $bhc;
print "$Pat BANNED messages.......$P2: $bnc\n" if $bnc;
print "$Pat SPAM messages.........$P2: $sc\n" if $sc;
print "$Pat CLEAN messages........$P2: $clc\n" if $clc;
print "$Pat messages (TOTAL)......$P2: " . ($ic + $bhc + $bnc + $sc +$clc) . "\n\n\n";
print "Legend\n";
print "------\n";
if($bhc and not $clc) {
print "Blocked BANNED messages.......: see comment at the end of this report\n";
}
print "INFECTED/BANNED columns are...: Date|Sender [reason] -> Recipient(s)\n";
print "All other columns are.........: Date|Score|DSN|Sender -> Recipient(s)\n";
print "Delivery Status Notification..: '+'= DSN sent, '-'= DSN not sent\n";
print "                                (INFECTED/BANNED: DSN always sent)\n\n";
if($ic) {
    print "\n$Pat INFECTED messages\n";
    print "$P3------------------------\n";
    for($c = 1; $c <= $ic; $c++) { print "$i[$c]\n"; };
}
if($bnc) {
    print "\n$Pat BANNED messages\n";
    print "$P3----------------------\n";
    for($c = 1; $c <= $bnc; $c++) { print "$bn[$c]\n"; };
}
if($bhc) {
    print "\n$Pat BAD-HEADER messages\n";
    print "$P3--------------------------\n";
    for($c = 1; $c <= $bhc; $c++) { print "$bh[$c]\n"; };
}
if($sc) {
    print "\n$Pat SPAM messages\n";
    print "$P3--------------------\n";
    for($c = 1; $c <= $sc; $c++) { print "$s[$c]\n"; };
}
if($clc) {
    print "\n$Pat CLEAN messages\n";
    print "$P3---------------------\n";
    for($c = 1; $c <= $clc; $c++) { print "$cl[$c]\n"; };
}
#if($bhc and not $clc) {
#    print "\n\n-------------------------------------------------------------------------------\n";
#    print " At the moment 'BAD-HEADER' messages basically are SPAM.\n";
#    print " The header checking code is not fully implemented yet, so these messages\n";
#    print " will always be passed to the recipient, unless the SPAM score is high enough\n";
#    print " to block them anyway.  That's when you'll find a blocked BAD-HEADER message\n";
#    print " in the report.\n";
#    print " Once the code is finished, these messages will be treated like INFECTED\n";
#    print " or BANNED messages, meaning, they will be blocked and a DSN will be sent\n";
#    print " to the original sender.\n";
#    print " The reason for that is, that BAD-HEADER messages can be used to break a virus\n";
#    print " into multiple emails and reassemble it in the recipient's inbox!\n";
#    print "-------------------------------------------------------------------------------\n";
#}
